#include <SocketHandler.h>
#include <TcpSocket.h>
#include <StdoutLog.h>
#include <signal.h>
#include <IEventOwner.h>
#include <EventHandler.h>

#define PORT 4040
#define TIME_US 250000

#define DEB(x) x


static	bool quit = false;
static	EventHandler *ph = NULL;


void sighandler(int s)
{
	quit = true;
	ph -> SetQuit();
}


void sigpipe(int s)
{
}


class evHandler : public EventHandler
{
public:
	evHandler() : EventHandler() {}
	evHandler(StdLog *p) : EventHandler(p) {}
};


class lSocket : public TcpSocket, public IEventOwner
{
public:
	lSocket(ISocketHandler& h, bool first = false) : TcpSocket(h), IEventOwner(static_cast<evHandler&>(h)), m_id(0), m_first(first) {
		SetLineProtocol();
	}
	~lSocket() {
	}

	void OnConnect() {
		if (!m_id)
		{
			m_id = AddEvent(0, TIME_US);
		}
	}

	void OnDelete() {
		if (m_first)
			printf("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
				"@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
				"@@@@@@@@@@@@@@ OnDelete() first @@@@@@@@@@@@@@@@@@\n"
				"@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
				"@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
	}

	void OnEvent(int id) {
		if (id != m_id)
		{
			printf("ID MISMATCH!\n");
			exit(-1);
		}
		if (!quit)
		{
			lSocket *p = new lSocket(Handler());
			p -> SetDeleteByHandler();
			p -> Open("localhost", PORT);
			Handler().Add(p);
		}
		static_cast<evHandler&>(Handler()).SetQuit( quit );
		SetCloseAndDelete();
	}

private:
	int m_id;
	bool m_first;
};


class l2Socket : public TcpSocket, public IEventOwner
{
public:
	l2Socket(ISocketHandler& h, bool first = false) : TcpSocket(h), IEventOwner(static_cast<evHandler&>(h)), m_id(0), m_first(first) {
		SetLineProtocol();
	}
	~l2Socket() {
	}

	void OnConnect() {
	}

	void OnLine(const std::string& line) {
		if (!m_id)
		{
			m_id = AddEvent(0, TIME_US);
		}
	}

	void OnDelete() {
		if (m_first)
			printf("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
				"@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
				"@@@@@@@@@@@@@@ OnDelete() first @@@@@@@@@@@@@@@@@@\n"
				"@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
				"@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
	}

	void OnEvent(int id) {
		if (id != m_id)
		{
			printf("ID MISMATCH!\n");
			exit(-1);
		}
		if (!quit)
		{
			l2Socket *p = new l2Socket(Handler());
			p -> SetDeleteByHandler();
			p -> Open("localhost", PORT);
			Handler().Add(p);
		}
		static_cast<evHandler&>(Handler()).SetQuit( quit );
		SetCloseAndDelete();
	}

private:
	int m_id;
	bool m_first;
};


class l3Socket : public TcpSocket, public IEventOwner
{
public:
	l3Socket(ISocketHandler& h, bool first = false) : TcpSocket(h), IEventOwner(static_cast<evHandler&>(h)), m_ehlo(false), m_id(0), m_first(first) {
		SetLineProtocol();
	}
	~l3Socket() {
	}

	void OnConnect() {
	}

	void OnLine(const std::string& line) {
		if (!m_ehlo)
		{
			Send("EHLO alhem.net\r\n");
			m_ehlo = true;
		}
		else
		if (!m_id)
		{
			m_id = AddEvent(0, TIME_US);
		}
	}

	void OnDelete() {
		if (m_first)
			printf("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
				"@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
				"@@@@@@@@@@@@@@ OnDelete() first @@@@@@@@@@@@@@@@@@\n"
				"@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
				"@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
	}
	
	void OnEvent(int id) {
		if (id != m_id)
		{
			printf("ID MISMATCH!\n");
			exit(-1);
		}
		if (!quit)
		{
			l3Socket *p = new l3Socket(Handler());
			p -> SetDeleteByHandler();
			p -> Open("localhost", PORT);
			Handler().Add(p);
		}
		static_cast<evHandler&>(Handler()).SetQuit( quit );
		Send("QUIT\r\n");
	}

private:
	bool m_ehlo;
	int m_id;
	bool m_first;
};


int main(int argc, char *argv[])
{
	signal(SIGHUP, sighandler);
	signal(SIGINT, sighandler);
	signal(SIGPIPE, sigpipe);

	{
		StdoutLog log;
		{
			evHandler h(&log);
			ph = &h;
			{
				lSocket sock(h, true);
				sock.Open("localhost", PORT);
				h.Add(&sock);
				{
					l2Socket sock2(h, true);
					sock2.Open("localhost", PORT);
					h.Add(&sock2);
					{
						l3Socket sock3(h, true);
						sock3.Open("localhost", PORT);
						h.Add(&sock3);

						h.EventLoop();

						int count = 10;
						while (count--)
						{
							printf("Timeout %d\n", count);
							fflush(stdout);
							h.Select(1, 0);
						}
						printf("Exiting 1...\n");
						fflush(stdout);
					}
					printf("Exiting 2...\n");
					fflush(stdout);
				}
				printf("Exiting 3...\n");
				fflush(stdout);
			}
			printf("Exiting 4...\n");
			fflush(stdout);
		}
		printf("Exiting 5...\n");
		fflush(stdout);
	}
	printf("Exiting 6...\n");
	fflush(stdout);
}


